==> qsiprep-1.0.0rc1.tar.gz: qsiprep-1.0.0rc1/qsiprep/interfaces/freesurfer.py <==
#!/usr/bin/env python
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
FreeSurfer tools interfaces
~~~~~~~~~~~~~~~~~~~~~~~~~~~

Fetch some example data:

    >>> import os
    >>> from niworkflows import data
    >>> data_root = data.get_bids_examples(variant='BIDS-examples-1-enh-ds054')
    >>> os.chdir(data_root)

Disable warnings:

    >>> from nipype import logging
    >>> logging.getLogger('nipype.interface').setLevel('ERROR')

"""
import os.path as op
from pathlib import Path

import nibabel as nb
import numpy as np
from nilearn.image import new_img_like, resample_to_img
from nipype.interfaces import freesurfer as fs
from nipype.interfaces.afni import Zeropad
from nipype.interfaces.base import (
    BaseInterfaceInputSpec,
    Directory,
    File,
    InputMultiPath,
    SimpleInterface,
    TraitedSpec,
    isdefined,
    traits,
)
from nipype.interfaces.freesurfer.base import FSCommandOpenMP, FSTraitedSpec
from nipype.interfaces.freesurfer.utils import LTAConvert
from nipype.utils.filemanip import copyfile, filename_to_list, fname_presuffix
from niworkflows.utils.images import _copyxform
from scipy import ndimage
from scipy.ndimage.morphology import binary_fill_holes
from skimage import morphology as sim


class FSTraitedSpecOpenMP(FSTraitedSpec):
    num_threads = traits.Int(desc="allows for specifying more threads", nohash=True)


class StructuralReference(fs.RobustTemplate):
    """Variation on RobustTemplate that simply copies the source if a single
    volume is provided.

    >>> from qsiprep.utils.bids import collect_data
    >>> t1w = collect_data('ds114', '01')[0]['t1w']
    >>> template = StructuralReference()
    >>> template.inputs.in_files = t1w
    >>> template.inputs.auto_detect_sensitivity = True
    >>> template.cmdline  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    'mri_robust_template --satit --mov .../sub-01_ses-retest_T1w.nii.gz
        .../sub-01_ses-test_T1w.nii.gz --template mri_robust_template_out.mgz'

    """

    def _num_vols(self):
        n_files = len(self.inputs.in_files)
        if n_files != 1:
            return n_files

        img = nb.load(self.inputs.in_files[0])
        if len(img.shape) == 3:
            return 1

        return img.shape[3]

    @property
    def cmdline(self):
        if self._num_vols() == 1:
            return "echo Only one time point!"
        return super(StructuralReference, self).cmdline

    def _list_outputs(self):
        outputs = super(StructuralReference, self)._list_outputs()
        if self._num_vols() == 1:
            in_file = self.inputs.in_files[0]
            outputs["out_file"] = in_file
            if isdefined(outputs["transform_outputs"]):
                transform_file = outputs["transform_outputs"][0]
                fs.utils.LTAConvert(
                    in_lta="identity.nofile",
                    source_file=in_file,
                    target_file=in_file,
                    out_lta=transform_file,
                ).run()
        return outputs


class MakeMidthicknessInputSpec(fs.utils.MRIsExpandInputSpec):
    graymid = InputMultiPath(desc="Existing graymid/midthickness file")


class MakeMidthickness(fs.MRIsExpand):
    """Variation on MRIsExpand that checks for an existing midthickness/graymid
    surface, and copies if available.

    mris_expand is an expensive operation, so this avoids re-running it when the
    working directory is lost.
    If users provide their own midthickness/graymid file, we assume they have
    created it correctly.
    """

    input_spec = MakeMidthicknessInputSpec

    @property
    def cmdline(self):
        cmd = super(MakeMidthickness, self).cmdline
        if not isdefined(self.inputs.graymid) or len(self.inputs.graymid) < 1:
            return cmd

        # Possible graymid values inclue {l,r}h.{graymid,midthickness}
        # Prefer midthickness to graymid, require to be of the same hemisphere
        # as input
        source = None
        in_base = op.basename(self.inputs.in_file)
        mt = self._associated_file(in_base, "midthickness")
        gm = self._associated_file(in_base, "graymid")

        for surf in self.inputs.graymid:
            if op.basename(surf) == mt:
                source = surf
                break
            if op.basename(surf) == gm:
                source = surf

        if source is None:
            return cmd

        return "cp {} {}".format(source, self._list_outputs()["out_file"])


class FSInjectBrainExtractedInputSpec(BaseInterfaceInputSpec):
    subjects_dir = Directory(mandatory=True, desc="FreeSurfer SUBJECTS_DIR")
    subject_id = traits.Str(mandatory=True, desc="Subject ID")
    in_brain = File(mandatory=True, exists=True, desc="input file, part of a BIDS tree")


class FSInjectBrainExtractedOutputSpec(TraitedSpec):
    subjects_dir = Directory(desc="FreeSurfer SUBJECTS_DIR")
    subject_id = traits.Str(desc="Subject ID")


class FSInjectBrainExtracted(SimpleInterface):
    input_spec = FSInjectBrainExtractedInputSpec
    output_spec = FSInjectBrainExtractedOutputSpec
    _always_run = True

    def _run_interface(self, runtime):
        subjects_dir, subject_id = inject_skullstripped(
            self.inputs.subjects_dir, self.inputs.subject_id, self.inputs.in_brain
        )
        self._results["subjects_dir"] = subjects_dir
        self._results["subject_id"] = subject_id
        return runtime


class FSDetectInputsInputSpec(BaseInterfaceInputSpec):
    t1w_list = InputMultiPath(
        File(exists=True), mandatory=True, desc="input file, part of a BIDS tree"
    )
    t2w_list = InputMultiPath(File(exists=True), desc="input file, part of a BIDS tree")
    flair_list = InputMultiPath(File(exists=True), desc="input file, part of a BIDS tree")
    hires_enabled = traits.Bool(True, usedefault=True, desc="enable hi-resolution processing")


class FSDetectInputsOutputSpec(TraitedSpec):
    t2w = File(desc="reference T2w image")
    use_t2w = traits.Bool(desc="enable use of T2w downstream computation")
    flair = File(desc="reference FLAIR image")
    use_flair = traits.Bool(desc="enable use of FLAIR downstream computation")
    hires = traits.Bool(desc="enable hi-res processing")
    mris_inflate = traits.Str(desc="mris_inflate argument")


class FSDetectInputs(SimpleInterface):
    input_spec = FSDetectInputsInputSpec
    output_spec = FSDetectInputsOutputSpec

    def _run_interface(self, runtime):
        t2w, flair, self._results["hires"], mris_inflate = detect_inputs(
            self.inputs.t1w_list,
            t2w_list=self.inputs.t2w_list if isdefined(self.inputs.t2w_list) else None,
            flair_list=self.inputs.flair_list if isdefined(self.inputs.flair_list) else None,
            hires_enabled=self.inputs.hires_enabled,
        )

        self._results["use_t2w"] = t2w is not None
        if self._results["use_t2w"]:
            self._results["t2w"] = t2w

        self._results["use_flair"] = flair is not None
        if self._results["use_flair"]:
            self._results["flair"] = flair

        if self._results["hires"]:
            self._results["mris_inflate"] = mris_inflate

        return runtime


class TruncateLTA(object):
    """Mixin to ensure that LTA files do not store overly long paths,
    which lead to segmentation faults when read by FreeSurfer tools.

    See the following issues for discussion:

    * https://github.com/freesurfer/freesurfer/pull/180
    * https://github.com/poldracklab/fmriprep/issues/768
    * https://github.com/poldracklab/fmriprep/pull/778
    * https://github.com/poldracklab/fmriprep/issues/1268
    * https://github.com/poldracklab/fmriprep/pull/1274
    """

    # Use a tuple in case some object produces multiple transforms
    lta_outputs = ("out_lta_file",)

    def _post_run_hook(self, runtime):

        outputs = self._list_outputs()

        for lta_name in self.lta_outputs:
            lta_file = outputs[lta_name]
            if not isdefined(lta_file):
                continue

            with open(lta_file, "r") as f:
                lines = f.readlines()

            fixed = False
            newfile = []

            for line in lines:
                if line.startswith("filename = ") and len(line.strip("\n")) >= 255:
                    fixed = True
                    newfile.append("filename = path_too_long\n")
                else:
                    newfile.append(line)

            if fixed:
                with open(lta_file, "w") as f:
                    f.write("".join(newfile))

        runtime = super(TruncateLTA, self)._post_run_hook(runtime)

        return runtime


class PatchedLTAConvert(TruncateLTA, LTAConvert):
    """
    LTAconvert is producing a lta file refer as out_lta
    truncate filename through mixin TruncateLTA
    """

    lta_outputs = ("out_lta",)


class RefineBrainMaskInputSpec(BaseInterfaceInputSpec):
    in_anat = File(exists=True, mandatory=True, desc="input anatomical reference (INU corrected)")
    in_aseg = File(exists=True, mandatory=True, desc="input ``aseg`` file, in NifTi format.")
    in_ants = File(
        exists=True,
        mandatory=True,
        desc="brain tissue segmentation generated with antsBrainExtraction.sh",
    )


class RefineBrainMaskOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="new mask")


class RefineBrainMask(SimpleInterface):
    """
    Refine the brain mask implicit in the ``aseg.mgz``
    file to include possibly missing gray-matter voxels
    and deep, wide sulci.
    """

    input_spec = RefineBrainMaskInputSpec
    output_spec = RefineBrainMaskOutputSpec

    def _run_interface(self, runtime):

        self._results["out_file"] = fname_presuffix(
            self.inputs.in_anat, suffix="_rbrainmask", newpath=runtime.cwd
        )

        anatnii = nb.load(self.inputs.in_anat)
        msknii = nb.Nifti1Image(
            grow_mask(
                anatnii.get_fdata(),
                nb.load(self.inputs.in_aseg).get_fdata(),
                nb.load(self.inputs.in_ants).get_fdata(),
            ),
            anatnii.affine,
            anatnii.header,
        )
        msknii.set_data_dtype(np.uint8)
        msknii.to_filename(self._results["out_file"])

        return runtime


def inject_skullstripped(subjects_dir, subject_id, skullstripped):
    mridir = op.join(subjects_dir, subject_id, "mri")
    t1 = op.join(mridir, "T1.mgz")
    bm_auto = op.join(mridir, "brainmask.auto.mgz")
    bm = op.join(mridir, "brainmask.mgz")

    if not op.exists(bm_auto):
        img = nb.load(t1)
        mask = nb.load(skullstripped)
        bmask = new_img_like(mask, mask.get_fdata() > 0)
        resampled_mask = resample_to_img(bmask, img, "nearest")
        masked_image = new_img_like(img, img.get_fdata() * resampled_mask.get_fdata())
        masked_image.to_filename(bm_auto)

    if not op.exists(bm):
        copyfile(bm_auto, bm, copy=True, use_hardlink=True)

    return subjects_dir, subject_id


def detect_inputs(t1w_list, t2w_list=None, flair_list=None, hires_enabled=True):
    t1w_list = filename_to_list(t1w_list)
    t2w_list = filename_to_list(t2w_list) if t2w_list is not None else []
    flair_list = filename_to_list(flair_list) if flair_list is not None else []
    t1w_ref = nb.load(t1w_list[0])
    # Use high resolution preprocessing if voxel size < 1.0mm
    # Tolerance of 0.05mm requires that rounds down to 0.9mm or lower
    hires = hires_enabled and max(t1w_ref.header.get_zooms()) < 1 - 0.05

    t2w = None
    if t2w_list and max(nb.load(t2w_list[0]).header.get_zooms()) < 1.2:
        t2w = t2w_list[0]

    # Prefer T2w to FLAIR if both present and T2w satisfies
    flair = None
    if flair_list and not t2w and max(nb.load(flair_list[0]).header.get_zooms()) < 1.2:
        flair = flair_list[0]

    # https://surfer.nmr.mgh.harvard.edu/fswiki/SubmillimeterRecon
    mris_inflate = "-n 50" if hires else None
    return (t2w, flair, hires, mris_inflate)


def refine_aseg(aseg, ball_size=4):
    """
    First step to reconcile ANTs' and FreeSurfer's brain masks.

    Here, the ``aseg.mgz`` mask from FreeSurfer is refined in two
    steps, using binary morphological operations:

      1. With a binary closing operation the sulci are included
         into the mask. This results in a smoother brain mask
         that does not exclude deep, wide sulci.

      2. Fill any holes (typically, there could be a hole next to
         the pineal gland and the corpora quadrigemina if the great
         cerebral brain is segmented out).


    """
    # Read aseg data
    bmask = aseg.copy()
    bmask[bmask > 0] = 1
    bmask = bmask.astype(np.uint8)

    # Morphological operations
    selem = sim.ball(ball_size)
    newmask = sim.binary_closing(bmask, selem)
    newmask = binary_fill_holes(newmask.astype(np.uint8), selem).astype(np.uint8)

    return newmask.astype(np.uint8)


def grow_mask(anat, aseg, ants_segs=None, ww=7, zval=2.0, bw=4):
    """
    Grow mask including pixels that have a high likelihood.
    GM tissue parameters are sampled in image patches of ``ww`` size.

    This is inspired on mindboggle's solution to the problem:
    https://github.com/nipy/mindboggle/blob/master/mindboggle/guts/segment.py#L1660

    """
    selem = sim.ball(bw)

    if ants_segs is None:
        ants_segs = np.zeros_like(aseg, dtype=np.uint8)

    aseg[aseg == 42] = 3  # Collapse both hemispheres
    gm = anat.copy()
    gm[aseg != 3] = 0

    refined = refine_aseg(aseg)
    newrefmask = sim.binary_dilation(refined, selem) - refined
    indices = np.argwhere(newrefmask > 0)
    for pixel in indices:
        # When ATROPOS identified the pixel as GM, set and carry on
        if ants_segs[tuple(pixel)] == 2:
            refined[tuple(pixel)] = 1
            continue

        window = gm[
            pixel[0] - ww : pixel[0] + ww,
            pixel[1] - ww : pixel[1] + ww,
            pixel[2] - ww : pixel[2] + ww,
        ]
        if np.any(window > 0):
            mu = window[window > 0].mean()
            sigma = max(window[window > 0].std(), 1.0e-5)
            zstat = abs(anat[tuple(pixel)] - mu) / sigma
            refined[tuple(pixel)] = int(zstat < zval)

    refined = sim.binary_opening(refined, selem)
    return refined


def medial_wall_to_nan(in_file, subjects_dir, target_subject, newpath=None):
    """Convert values on medial wall to NaNs"""
    import os

    import nibabel as nb
    import numpy as np

    fn = os.path.basename(in_file)
    if not target_subject.startswith("fs"):
        return in_file

    cortex = nb.freesurfer.read_label(
        os.path.join(subjects_dir, target_subject, "label", "{}.cortex.label".format(fn[:2]))
    )
    func = nb.load(in_file)
    medial = np.delete(np.arange(len(func.darrays[0].data)), cortex)
    for darray in func.darrays:
        darray.data[medial] = np.nan

    out_file = os.path.join(newpath or os.getcwd(), fn)
    func.to_filename(out_file)
    return out_file


def find_fs_path(freesurfer_dir, subject_id):
    if freesurfer_dir is None:
        return None
    nosub = op.join(freesurfer_dir, subject_id)
    if op.exists(nosub):
        return Path(nosub)
    withsub = op.join(freesurfer_dir, "sub-" + subject_id)
    if op.exists(withsub):
        return Path(withsub)
    return None


class _PrepareSynthStripGridInputSpec(BaseInterfaceInputSpec):
    input_image = File(exists=True, mandatory=True)


class _PrepareSynthStripGridOutputSpec(TraitedSpec):
    prepared_image = File(exists=True)


class PrepareSynthStripGrid(SimpleInterface):
    input_spec = _PrepareSynthStripGridInputSpec
    output_spec = _PrepareSynthStripGridOutputSpec

    def _run_interface(self, runtime):
        out_fname = fname_presuffix(
            self.inputs.input_image,
            newpath=runtime.cwd,
            suffix="_SynthStripGrid.nii",
            use_ext=False,
        )
        self._results["prepared_image"] = out_fname

        # possibly downsample the image for sloppy mode. Always ensure float32
        img = nb.load(self.inputs.input_image)
        if not img.ndim == 3:
            raise Exception("3D inputs are required for Synthstrip")
        xvoxels, yvoxels, zvoxels = img.shape

        def get_padding(nvoxels):
            extra_slices = nvoxels % 64
            if extra_slices == 0:
                return 0
            complete_64s = nvoxels // 64
            return 64 * (complete_64s + 1) - nvoxels

        def split_padding(padding):
            halfpad = padding // 2
            return halfpad, halfpad + halfpad % 2

        spad = get_padding(zvoxels)
        rpad, lpad = split_padding(get_padding(xvoxels))
        apad, ppad = split_padding(get_padding(yvoxels))

        zeropad = Zeropad(
            S=spad,
            R=rpad,
            L=lpad,
            A=apad,
            P=ppad,
            in_files=self.inputs.input_image,
            out_file=out_fname,
        )

        _ = zeropad.run()
        assert op.exists(out_fname)
        return runtime


class _SynthStripInputSpec(FSTraitedSpecOpenMP):
    input_image = File(argstr="-i %s", exists=True, mandatory=True)
    no_csf = traits.Bool(argstr="--no-csf", desc="Exclude CSF from brain border.")
    border = traits.Int(argstr="-b %d", desc="Mask border threshold in mm. Default is 1.")
    gpu = traits.Bool(argstr="-g")
    out_brain = File(
        argstr="-o %s",
        name_template="%s_brain.nii.gz",
        name_source=["input_image"],
        keep_extension=False,
        desc="skull stripped image with corrupt sform",
    )
    out_brain_mask = File(
        argstr="-m %s",
        name_template="%s_mask.nii.gz",
        name_source=["input_image"],
        keep_extension=False,
        desc="mask image with corrupt sform",
    )


class _SynthStripOutputSpec(TraitedSpec):
    out_brain = File(exists=True)
    out_brain_mask = File(exists=True)


class SynthStrip(FSCommandOpenMP):
    input_spec = _SynthStripInputSpec
    output_spec = _SynthStripOutputSpec
    _cmd = "mri_synthstrip"

    def _num_threads_update(self):
        if self.inputs.num_threads:
            self.inputs.environ.update({"OMP_NUM_THREADS": "1"})


class FixHeaderSynthStrip(SynthStrip):

    def _run_interface(self, runtime, correct_return_codes=(0,)):
        # Run normally
        runtime = super(FixHeaderSynthStrip, self)._run_interface(runtime, correct_return_codes)

        outputs = self._list_outputs()
        if not op.exists(outputs["out_brain"]):
            raise Exception("mri_synthstrip failed!")

        if outputs.get("out_brain_mask"):
            _copyxform(self.inputs.input_image, outputs["out_brain_mask"])

        _copyxform(self.inputs.input_image, outputs["out_brain"])

        return runtime


class MockSynthStrip(SimpleInterface):
    input_spec = _SynthStripInputSpec
    output_spec = _SynthStripOutputSpec

    def _run_interface(self, runtime):
        from nipype.interfaces.fsl import BET

        this_bet = BET(
            mask=True,
            in_file=self.inputs.input_image,
            output_type="NIFTI_GZ",
        )
        result = this_bet.run()
        self._results["out_brain"] = result.outputs.out_file
        self._results["out_brain_mask"] = result.outputs.mask_file

        return runtime


class _SynthSegInputSpec(FSTraitedSpecOpenMP):
    input_image = File(argstr="--i %s", exists=True, mandatory=True)
    num_threads = traits.Int(
        default=1, argstr="--threads %d", usedefault=True, desc="Number of threads to use"
    )
    fast = traits.Bool(argstr="--fast", desc="fast predictions (lower quality).")
    robust = traits.Bool(argstr="--robust", desc="use robust predictions (slower).")
    out_seg = File(
        argstr="--o %s",
        name_template="%s_aseg.nii.gz",
        name_source=["input_image"],
        keep_extension=False,
        desc="segmentation image",
    )
    out_post = File(
        argstr="--post %s",
        name_template="%s_post.nii.gz",
        name_source=["input_image"],
        keep_extension=False,
        desc="posteriors image",
    )
    out_qc = File(
        argstr="--qc %s",
        name_template="%s_qc.csv",
        name_source=["input_image"],
        keep_extension=False,
        desc="qc csv",
    )
    cpu = traits.Bool(
        True, argstr="--cpu", usedefault=True, desc="Enforce running with CPU rather than GPU."
    )


class _SynthSegOutputSpec(TraitedSpec):
    out_seg = File(exists=True)
    out_post = File(exists=True)
    out_qc = File(exists=True)


class SynthSeg(FSCommandOpenMP):
    input_spec = _SynthSegInputSpec
    output_spec = _SynthSegOutputSpec
    _cmd = "mri_synthseg"

    def _format_arg(self, name, trait_spec, value):
        # Hardcode threads to be 1
        if name == "num_threads":
            return "--threads 1"
        return super()._format_arg(name, trait_spec, value)

    def _num_threads_update(self):
        if self.inputs.num_threads:
            self.inputs.environ.update({"OMP_NUM_THREADS": "1"})


class MockSynthSeg(SimpleInterface):
    """A fake version of synthseg for testing."""

    input_spec = _SynthSegInputSpec
    output_spec = _SynthSegOutputSpec

    def _run_interface(self, runtime):
        from nipype.interfaces.fsl import BET

        output_qc = op.join(runtime.cwd, "fake_synthseg_qc.csv")
        with open(output_qc, "w") as qcf:
            qcf.write("Test QC file\n")

        # Get a brain mask
        this_bet = BET(
            mask=True,
            in_file=self.inputs.input_image,
            output_type="NIFTI_GZ",
        )
        result = this_bet.run()
        self._results["out_post"] = result.outputs.out_file

        # Make a fake segmentation
        img = nb.load(result.outputs.mask_file)
        orig_mask = img.get_fdata() > 0
        eroded1 = ndimage.binary_erosion(orig_mask, iterations=3)
        eroded2 = ndimage.binary_erosion(eroded1, iterations=3)
        final = orig_mask.astype(int) + eroded1 + eroded2
        out_img = nb.Nifti1Image(final, img.affine, header=img.header)
        out_fname = fname_presuffix(self.inputs.input_image, suffix="_dseg", newpath=runtime.cwd)
        out_img.to_filename(out_fname)
        self._results["out_seg"] = out_fname
        return runtime

==> qsiprep-1.0.1.tar.gz: qsiprep-1.0.1/qsiprep/interfaces/freesurfer.py <==
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
FreeSurfer tools interfaces
~~~~~~~~~~~~~~~~~~~~~~~~~~~

Fetch some example data:

    >>> import os
    >>> from niworkflows import data
    >>> data_root = data.get_bids_examples(variant='BIDS-examples-1-enh-ds054')
    >>> os.chdir(data_root)

Disable warnings:

    >>> from nipype import logging
    >>> logging.getLogger('nipype.interface').setLevel('ERROR')

"""

import os.path as op

import nibabel as nb
from nipype.interfaces import freesurfer as fs
from nipype.interfaces.afni import Zeropad
from nipype.interfaces.base import (
    BaseInterfaceInputSpec,
    File,
    SimpleInterface,
    TraitedSpec,
    isdefined,
    traits,
)
from nipype.interfaces.freesurfer.base import FSCommandOpenMP, FSTraitedSpec
from nipype.utils.filemanip import fname_presuffix
from niworkflows.utils.images import _copyxform
from scipy import ndimage


class FSTraitedSpecOpenMP(FSTraitedSpec):
    num_threads = traits.Int(desc='allows for specifying more threads', nohash=True)


class StructuralReference(fs.RobustTemplate):
    """Variation on RobustTemplate that simply copies the source if a single
    volume is provided.

    >>> from qsiprep.utils.bids import collect_data
    >>> t1w = collect_data('ds114', '01')[0]['t1w']
    >>> template = StructuralReference()
    >>> template.inputs.in_files = t1w
    >>> template.inputs.auto_detect_sensitivity = True
    >>> template.cmdline  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    'mri_robust_template --satit --mov .../sub-01_ses-retest_T1w.nii.gz
        .../sub-01_ses-test_T1w.nii.gz --template mri_robust_template_out.mgz'

    """

    def _num_vols(self):
        n_files = len(self.inputs.in_files)
        if n_files != 1:
            return n_files

        img = nb.load(self.inputs.in_files[0])
        if len(img.shape) == 3:
            return 1

        return img.shape[3]

    @property
    def cmdline(self):
        if self._num_vols() == 1:
            return 'echo Only one time point!'
        return super().cmdline

    def _list_outputs(self):
        outputs = super()._list_outputs()
        if self._num_vols() == 1:
            in_file = self.inputs.in_files[0]
            outputs['out_file'] = in_file
            if isdefined(outputs['transform_outputs']):
                transform_file = outputs['transform_outputs'][0]
                fs.utils.LTAConvert(
                    in_lta='identity.nofile',
                    source_file=in_file,
                    target_file=in_file,
                    out_lta=transform_file,
                ).run()
        return outputs


class _PrepareSynthStripGridInputSpec(BaseInterfaceInputSpec):
    input_image = File(exists=True, mandatory=True)


class _PrepareSynthStripGridOutputSpec(TraitedSpec):
    prepared_image = File(exists=True)


class PrepareSynthStripGrid(SimpleInterface):
    input_spec = _PrepareSynthStripGridInputSpec
    output_spec = _PrepareSynthStripGridOutputSpec

    def _run_interface(self, runtime):
        out_fname = fname_presuffix(
            self.inputs.input_image,
            newpath=runtime.cwd,
            suffix='_SynthStripGrid.nii',
            use_ext=False,
        )
        self._results['prepared_image'] = out_fname

        # possibly downsample the image for sloppy mode. Always ensure float32
        img = nb.load(self.inputs.input_image)
        if not img.ndim == 3:
            raise Exception('3D inputs are required for Synthstrip')
        xvoxels, yvoxels, zvoxels = img.shape

        def get_padding(nvoxels):
            extra_slices = nvoxels % 64
            if extra_slices == 0:
                return 0
            complete_64s = nvoxels // 64
            return 64 * (complete_64s + 1) - nvoxels

        def split_padding(padding):
            halfpad = padding // 2
            return halfpad, halfpad + halfpad % 2

        spad = get_padding(zvoxels)
        rpad, lpad = split_padding(get_padding(xvoxels))
        apad, ppad = split_padding(get_padding(yvoxels))

        zeropad = Zeropad(
            S=spad,
            R=rpad,
            L=lpad,
            A=apad,
            P=ppad,
            in_files=self.inputs.input_image,
            out_file=out_fname,
        )

        _ = zeropad.run()
        assert op.exists(out_fname)
        return runtime


class _SynthStripInputSpec(FSTraitedSpecOpenMP):
    input_image = File(argstr='-i %s', exists=True, mandatory=True)
    no_csf = traits.Bool(argstr='--no-csf', desc='Exclude CSF from brain border.')
    border = traits.Int(argstr='-b %d', desc='Mask border threshold in mm. Default is 1.')
    gpu = traits.Bool(argstr='-g')
    out_brain = File(
        argstr='-o %s',
        name_template='%s_brain.nii.gz',
        name_source=['input_image'],
        keep_extension=False,
        desc='skull stripped image with corrupt sform',
    )
    out_brain_mask = File(
        argstr='-m %s',
        name_template='%s_mask.nii.gz',
        name_source=['input_image'],
        keep_extension=False,
        desc='mask image with corrupt sform',
    )


class _SynthStripOutputSpec(TraitedSpec):
    out_brain = File(exists=True)
    out_brain_mask = File(exists=True)


class SynthStrip(FSCommandOpenMP):
    input_spec = _SynthStripInputSpec
    output_spec = _SynthStripOutputSpec
    _cmd = 'mri_synthstrip'

    def _num_threads_update(self):
        if self.inputs.num_threads:
            self.inputs.environ.update({'OMP_NUM_THREADS': '1'})


class FixHeaderSynthStrip(SynthStrip):
    def _run_interface(self, runtime, correct_return_codes=(0,)):
        # Run normally
        runtime = super()._run_interface(runtime, correct_return_codes)

        outputs = self._list_outputs()
        if not op.exists(outputs['out_brain']):
            raise Exception('mri_synthstrip failed!')

        if outputs.get('out_brain_mask'):
            _copyxform(self.inputs.input_image, outputs['out_brain_mask'])

        _copyxform(self.inputs.input_image, outputs['out_brain'])

        return runtime


class MockSynthStrip(SimpleInterface):
    input_spec = _SynthStripInputSpec
    output_spec = _SynthStripOutputSpec

    def _run_interface(self, runtime):
        from nipype.interfaces.fsl import BET

        this_bet = BET(
            mask=True,
            in_file=self.inputs.input_image,
            output_type='NIFTI_GZ',
        )
        result = this_bet.run()
        self._results['out_brain'] = result.outputs.out_file
        self._results['out_brain_mask'] = result.outputs.mask_file

        return runtime


class _SynthSegInputSpec(FSTraitedSpecOpenMP):
    input_image = File(argstr='--i %s', exists=True, mandatory=True)
    num_threads = traits.Int(
        default=1, argstr='--threads %d', usedefault=True, desc='Number of threads to use'
    )
    fast = traits.Bool(argstr='--fast', desc='fast predictions (lower quality).')
    robust = traits.Bool(argstr='--robust', desc='use robust predictions (slower).')
    out_seg = File(
        argstr='--o %s',
        name_template='%s_aseg.nii.gz',
        name_source=['input_image'],
        keep_extension=False,
        desc='segmentation image',
    )
    out_post = File(
        argstr='--post %s',
        name_template='%s_post.nii.gz',
        name_source=['input_image'],
        keep_extension=False,
        desc='posteriors image',
    )
    out_qc = File(
        argstr='--qc %s',
        name_template='%s_qc.csv',
        name_source=['input_image'],
        keep_extension=False,
        desc='qc csv',
    )
    cpu = traits.Bool(
        True, argstr='--cpu', usedefault=True, desc='Enforce running with CPU rather than GPU.'
    )


class _SynthSegOutputSpec(TraitedSpec):
    out_seg = File(exists=True)
    out_post = File(exists=True)
    out_qc = File(exists=True)


class SynthSeg(FSCommandOpenMP):
    input_spec = _SynthSegInputSpec
    output_spec = _SynthSegOutputSpec
    _cmd = 'mri_synthseg'

    def _format_arg(self, name, trait_spec, value):
        # Hardcode threads to be 1
        if name == 'num_threads':
            return '--threads 1'
        return super()._format_arg(name, trait_spec, value)

    def _num_threads_update(self):
        if self.inputs.num_threads:
            self.inputs.environ.update({'OMP_NUM_THREADS': '1'})


class MockSynthSeg(SimpleInterface):
    """A fake version of synthseg for testing."""

    input_spec = _SynthSegInputSpec
    output_spec = _SynthSegOutputSpec

    def _run_interface(self, runtime):
        from nipype.interfaces.fsl import BET

        output_qc = op.join(runtime.cwd, 'fake_synthseg_qc.csv')
        with open(output_qc, 'w') as qcf:
            qcf.write('Test QC file\n')

        # Get a brain mask
        this_bet = BET(
            mask=True,
            in_file=self.inputs.input_image,
            output_type='NIFTI_GZ',
        )
        result = this_bet.run()
        self._results['out_post'] = result.outputs.out_file

        # Make a fake segmentation
        img = nb.load(result.outputs.mask_file)
        orig_mask = img.get_fdata() > 0
        eroded1 = ndimage.binary_erosion(orig_mask, iterations=3)
        eroded2 = ndimage.binary_erosion(eroded1, iterations=3)
        final = orig_mask.astype(int) + eroded1 + eroded2
        out_img = nb.Nifti1Image(final, img.affine, header=img.header)
        out_fname = fname_presuffix(self.inputs.input_image, suffix='_dseg', newpath=runtime.cwd)
        out_img.to_filename(out_fname)
        self._results['out_seg'] = out_fname
        return runtime
